<?php

// 

error_reporting(E_ALL);
ini_set('display_errors', 1);

if (version_compare(phpversion(), '5.4.0', '<')) {
    die('Export supports PHP 5.4.0 or newer');
}

define('REQUEST_SCHEME', empty($_SERVER['HTTPS']) || $_SERVER['HTTPS'] === 'off' ? 'http' : 'https');
define('DS', DIRECTORY_SEPARATOR);
define('PATH_TO_ROOT', '../../../../..');
define('ROOT', getenv('HTML_APP_ROOT') ? getenv('HTML_APP_ROOT') : realpath(PATH_TO_ROOT));
define('HTTP_ROOT', REQUEST_SCHEME . '://' . $_SERVER['HTTP_HOST'] . dirname($_SERVER['SCRIPT_NAME']) . '/' . PATH_TO_ROOT);

function writeFile($path, $content) {
    @mkdir(dirname($path), 0777, true);
    return file_put_contents($path, $content);
}

function unpackFso($fso, $path) {
    if (!is_array($fso['items'])) {
        return;
    }

    if (!file_exists($path)) {
        mkdir($path, 0777, true);
    }

    if (!is_writable($path)) {
        throw new Exception('Permission denied: ' . $path);
    }

    foreach ($fso['items'] as $name => $file) {
        if (isset($file['content']) && isset($file['type'])) {
            if ($file['content'] === '[DELETED]') {
                @unlink($path . DS . $name);
            } else {
                switch ($file['type']) {
                    case 'text':
                        file_put_contents($path . DS . $name, $file['content']);
                        break;
                    case 'data':
                        file_put_contents($path . DS . $name, base64_decode($file['content']));
                        break;
                }
            }
        } else if (isset($file['items']) && isset($file['type'])) {
            unpackFso($file, $path . DS . $name);
        }
    }
}

function packFso($path) {
    $result = array();

    if (is_file($path)) {
        $content = file_get_contents($path);

        if ($content === false) {
            throw new Exception('Permission denied: ' . $path);
        }
        $type = 'text';
        $ext = pathinfo($path, PATHINFO_EXTENSION);

        if (in_array($ext, ['jpg', 'jpeg', 'bmp', 'png', 'gif', 'svg'])) {
            $type = 'data';
            $content = base64_encode($content);
        }

        $result = array('type' => $type, 'content' => $content);
    } else if (is_dir($path)) {
        $result = array('type' => 'dir', 'items' => array());

        if ($d = opendir($path)) {
            while (($name = readdir($d)) !== false) {
                if (in_array($name, array('.', '..'))) {
                    continue;
                }

                $result['items'][$name] = packFso($path . DS . $name);
            }
            closedir($d);
        }
    }

    return $result;
}

function uploadChunk($uploadPath)
{
    $result = array();

    $contentRange = getHeader('Content-Range');
    $isLast = getParam('last');

    if (!isset($_FILES['chunk']) || !file_exists($_FILES['chunk']['tmp_name'])) {
        $result = array(
            'status' => 'error',
            'message' => 'Empty chunk data'
        );
    } else if (!$contentRange && !$isLast) {
        $result = array(
            'status' => 'error',
            'message' => 'Empty Content-Range header'
        );
    } else {
        $rangeBegin = 0;

        if ($contentRange) {
            $contentRange = str_replace('bytes ', '', $contentRange);
            list($range, $total) = explode('/', $contentRange);
            list($rangeBegin, $rangeEnd) = explode('-', $range);
        }

        $tmpPath = $uploadPath . '.upload';
        @mkdir(dirname($tmpPath), 0777, true);

        $f = fopen($tmpPath, 'c');

        if (flock($f, LOCK_EX)) {
            fseek($f, (int) $rangeBegin);
            fwrite($f, file_get_contents($_FILES['chunk']['tmp_name']));

            flock($f, LOCK_UN);
            fclose($f);
        } else {
            throw new Exception('Permission denied: ' . $tmpPath);
        }

        if ($isLast) {
            if (file_exists($uploadPath)) {
                @unlink($uploadPath);
            }

            @mkdir(dirname($uploadPath), 0777, true);
            if (!is_writable(dirname($uploadPath))) {
                throw new Exception('Permission denied: ' . $uploadPath);
            }

            if (!is_writable($tmpPath)) {
                throw new Exception('Permission denied: ' . $tmpPath);
            }

            rename($tmpPath, $uploadPath);

            $result = array(
                'status' => 'done'
            );
        } else {
            $result['status'] = 'processed';
        }
    }

    return $result;
}

function performGlob($mask, $flags)
{
    $files = glob($mask, $flags);

    if (!is_array($files)) {
        $files = array();
    }

    $folders = glob(dirname($mask) . '/*', GLOB_ONLYDIR | GLOB_NOSORT);

    if (!is_array($folders)) {
        $folders = array();
    }

    foreach ($folders as $dir) {
        $files = array_merge($files, performGlob($dir . '/' . basename($mask), $flags));
    }

    return $files;
}

function enumerateDir($dir, $filter = null, $option = RecursiveIteratorIterator::SELF_FIRST) {
    $list = [];

    if (!$filter) {
        $filter = function () {
            return true;
        };
    }

    $iterator = new RecursiveIteratorIterator(
        new RecursiveCallbackFilterIterator(
            new RecursiveDirectoryIterator($dir, RecursiveDirectoryIterator::SKIP_DOTS | FilesystemIterator::UNIX_PATHS),
            $filter
        ),
        $option
    );

    foreach ($iterator as $item) {
        $f = new stdClass;
        $f->isDir = $item->isDir();
        $f->subPath = $iterator->getSubPath();
        $f->subPathName = $iterator->getSubPathName();
        $f->realPath = $item->getRealPath();
        $f->fileName = $item->getFilename();
        $list[] = $f;
    }

    $iterator = null;

    return $list;
}

function rm_r($dir, $exclude = [], $deleteRoot = false)
{
    if (!is_dir($dir)) {
        @unlink($dir);
        return;
    }

    if (!file_exists($dir)) {
        return false;
    }

    $list = enumerateDir($dir, function ($current) use ($exclude) {
        return !in_array($current->getFilename(), $exclude);
    }, RecursiveIteratorIterator::CHILD_FIRST);

    foreach ($list as $item) {
        if ($item->isDir) {
            @rmdir($item->realPath);
        } else {
            @unlink($item->realPath);
        }
    }

    if ($deleteRoot) {
        return @rmdir($dir);
    }

    return true;
}

function cp_r($source, $dest, $exclude = [])
{
    @mkdir($dest, 0777, true);

    $list = enumerateDir($source, function ($current) use ($exclude) {
        return !in_array($current->getFilename(), $exclude);
    });

    foreach ($list as $item) {
        if ($item->isDir) {
            @mkdir($dest . DS . $item->subPathName, 0777);
        } else {
            @copy($item->realPath, $dest . DS . $item->subPathName);
        }
    }
}

function checkDirPermissions($path)
{
    return is_dir($path) && is_writable($path) && is_readable($path);
}

function checkFilePermissions($path)
{
    return file_exists($path) && is_writable($path) && is_readable($path);
}

function dispatch()
{
    $action = empty($_REQUEST['action']) ? '' : preg_replace('/[^a-z]/i', '', $_REQUEST['action']);

    if (function_exists($action . 'Action')) {
        try {
            $result = call_user_func($action . 'Action');
        } catch (PermissionException $e) {
            $result = [
                'status' => 'error',
                'result' => 'error',
                'type' => 'permission',
                'message' => $e->getMessage(),
                'stack' => $e->getTraceAsString()
            ];
        } catch (Exception $e) {
            $result = [
                'status' => 'error',
                'result' => 'error',
                'message' => $e->getMessage(),
                'stack' => $e->getTraceAsString()
            ];
        }
    } else {
        $result = ['status' => 'error', 'message' => 'no action'];
    }

    echo is_array($result) ? json_encode($result) : $result;
}

function getParam($name, $default = null)
{
    return empty($_REQUEST[$name]) ? $default : $_REQUEST[$name];
}

function getPost($name, $default = null)
{
    return empty($_POST[$name]) ? $default : $_POST[$name];
}

function getHeader($header)
{
    $temp = 'HTTP_' . strtoupper(str_replace('-', '_', $header));
    if (isset($_SERVER[$temp])) {
        return $_SERVER[$temp];
    }

    return null;
}

function getAvailableThemeName($path, $themeName)
{
    while (file_exists($path . DS . $themeName)) {
        preg_match('#(.*?)(\d{0,4})$#', $themeName, $m);
        $themeName = $m[1];
        $suffix = (int) $m[2];
        $suffix++;
        $themeName .= $suffix;
    };

    return $themeName;
}

function startsWith($haystack, $needle) {
    return $needle === "" || strrpos($haystack, $needle, -strlen($haystack)) !== FALSE;
}
function endsWith($haystack, $needle) {
    return $needle === "" || (($temp = strlen($haystack) - strlen($needle)) >= 0 && strpos($haystack, $needle, $temp) !== FALSE);
}

function prepareToDownload($target) {
    $exclude = ["result", "__editor__"];
    $resPath = THEME_ROOT . "/result";
    rm_r($resPath);
    $dnnResult = ManifestHelper::Create(THEME_ROOT, $target, $exclude);
    $files = $dnnResult->files;

    $list = enumerateDir(THEME_ROOT, function ($current) use ($exclude) {
        return !in_array($current->getFileName(), $exclude);
    });
    foreach ($list as $item) {
        if (!$item->isDir && in_array(str_replace("\\", "/", $item->realPath), str_replace("\\", "/", $files))) {

            if ($item->fileName == "manifest.dnn") {
                writeFile($resPath . "/" . $target . ".dnn", $dnnResult->manifest);
            } else {
                $content = file_get_contents($item->realPath);

                if (endsWith($item->fileName, ".ascx")) {
                    $content = preg_replace_callback('/url\(([^\)]+)\)/', function($matches) use ($files) {

                        if (array_filter($files, function($current) use ($matches) {
                            if(endsWith(str_replace("\\", "/", $current), "/assets/images/" . $matches[1])) {
                                return true;
                            }
                            return false;
                        })) {
                            return "<%= SkinPath %>assets/images/" . $matches[1];
                        }

                        return $matches[1];
                    }, $content);
                }

                $content = preg_replace('/<!--REMOVE-->[\w\W]*?<!--END REMOVE-->/', '', $content);
                writeFile($resPath . "/" . $item->subPathName, $content);
            }
        }
    }

    return ['status' => 'done', 'path' => $resPath];
}

class ManifestHelper
{
    public static function Create($source, $name, $exclude)
    {
        $filteredFiles = [];

        $manifest = enumerateDir($source,
            function($current) use (&$filteredFiles) {
                if (strtolower($current->getFilename()) == "manifest.dnn") {
                    $filteredFiles[] = $current->getRealPath();

                    return true;
                }
                return false;
            });

        $containerFiles = enumerateDir($source,
            function($current) use (&$filteredFiles) {
                if (preg_match("/module_\d+\.ascx/i", $current->getFilename()) > 0) {
                    $filteredFiles[] = $current->getRealPath();

                    return true;
                }
                return false;
            });

        $menuFolders = enumerateDir($source,
            function($current) {
                $fixedPath = str_replace("/", "\\", $current);
                return preg_match("/\\menu_\d+$/i", $fixedPath) > 0;
            });

        if (count($manifest) == 1)
        {
            $xml = simplexml_load_file($manifest[0]->realPath);

            $xml->packages[0]->package[0]["name"] = $name;

            $skinNode = $xml->xpath("/dotnetnuke/packages/package/components/component[@type='Skin']")[0];
            $skinFilesNode = $skinNode->skinFiles;

            $skinFilesNode->skinName = $name;

            // Menu files
            foreach ($menuFolders as $menuFolder) {
                $menuFiles = enumerateDir($menuFolder->realPath);

                foreach($menuFiles as $menuFile) {
                    $filteredFiles[] = $menuFile->realPath;

                    $menuFileNode = $skinFilesNode->addChild("skinFile");
                    $menuFileNode->path = $menuFolder->subPathName;
                    $menuFileNode->name = $menuFile->fileName;
                }
            }

            $containersNode = $xml->xpath("/dotnetnuke/packages/package/components/component[@type='Container']")[0];
            $containersFilesNode = $containersNode->containerFiles;
            $containersFilesNode->containerName = $name;

            foreach ($containerFiles as $containerFile) {
                $containersFilesNode->addChild("containerFile")->name = $containerFile->fileName;
            }

            $skinFiles = enumerateDir($source, function($current) use ($filteredFiles, $exclude) {
                return !in_array($current->getRealPath(), $filteredFiles) && !in_array($current->getFilename(), $exclude);
            });

            foreach ($skinFiles as $skinFile) {

                if ($skinFile->isDir)
                    continue;

                if (startsWith($skinFile->subPathName, "assets")) {
                    $skinFileNode = $skinFilesNode->addChild("skinFile");

                    $skinFileNode->path = $skinFile->subPath;
                    $skinFileNode->name = $skinFile->fileName;

                    $filteredFiles[] =  $skinFile->realPath;
                }

                if (endsWith($skinFile->fileName, "ascx")) {
                    $skinFileNode = $skinFilesNode->addChild("skinFile");
                    $skinFileNode->name = $skinFile->fileName;
                    $filteredFiles[] =  $skinFile->realPath;

                    $skinFileNode = $skinFilesNode->addChild("skinFile");
                    $skinFileNode->name = str_replace(".ascx", ".doctype.xml", $skinFile->fileName);
                    $filteredFiles[] =  str_replace(".ascx", ".doctype.xml", $skinFile->realPath);
                }
            }
			
            $dom = new DOMDocument('1.0');
            $dom->preserveWhiteSpace = false;
            $dom->formatOutput = true;
            $dom->loadXML($xml->asXML());

            $result = new stdClass;
            $result->files = $filteredFiles;
            $result->manifest = $dom->saveXML();

            return $result;
        }

        throw new ErrorException("Unable to find manifest file.");

    }
}

class ZipHelper
{
    public static function zip($source, $wrapDir = '', $exclude = []) {
        if ($wrapDir && substr($wrapDir, -1) !== '/')
            $wrapDir .= '/';

        $outPath = tempnam('tmp', 'zip');
        $z = new ZipArchive();
        $res = $z->open($outPath, ZipArchive::CREATE);

        if ($res === true) {
            $list = enumerateDir($source, function ($current) use ($exclude) {
                return !in_array($current->getFilename(), $exclude);
            });

            foreach ($list as $item) {
                if (!$item->isDir) {
                    $z->addFile($item->realPath, $wrapDir . $item->subPathName);
                }
            }

            $z->close();
        } else {
            return ['status' => 'error', 'message' => self::message($res)];
        }

        return ['status' => 'done', 'path' => $outPath];
    }

    public static function unzip($file, $target) {
        $zip = new ZipArchive;
        $res = $zip->open($file);
        @mkdir($target, 0777, true);

        if ($res === true) {
            for($i = 0; $i < $zip->numFiles; $i++) {
                $data = $zip->getFromIndex($i);
                $filename = preg_replace('#[/\\\]#', DS, $zip->getNameIndex($i));
                $dest = $target . DS . $filename;

                if (substr($dest, -1) !== DS) {
                    @mkdir(dirname($dest), 0777, true);
                    file_put_contents($dest, $data);
                }
            }

            $zip->close();
            return ['status' => 'done'];
        } else {
            return ['status' => 'error', 'message' => self::message($res)];
        }
    }

    public static function unzipString($str, $key) {
        $zipPath = tempnam('tmp', 'str_zip');
        file_put_contents($zipPath, $str);

        $unzipPath = getcwd() . DS . 'unzip_string';
        @mkdir($unzipPath, 0777, true);

        $result = self::unzip($zipPath, $unzipPath);
        if ($result['status'] === 'done' && file_exists($unzipPath . DS . $key)) {
            $result['data'] = file_get_contents($unzipPath . DS . $key);
        } else {
            $result['message'] = 'unzip error';
        }

        rm_r($unzipPath, [], true);

        return $result;
    }

    public static function message($code)
    {
        switch ($code)
        {
            case 0:
                return 'No error';

            case 1:
                return 'Multi-disk zip archives not supported';

            case 2:
                return 'Renaming temporary file failed';

            case 3:
                return 'Closing zip archive failed';

            case 4:
                return 'Seek error';

            case 5:
                return 'Read error';

            case 6:
                return 'Write error';

            case 7:
                return 'CRC error';

            case 8:
                return 'Containing zip archive was closed';

            case 9:
                return 'No such file';

            case 10:
                return 'File already exists';

            case 11:
                return 'Can\'t open file';

            case 12:
                return 'Failure to create temporary file';

            case 13:
                return 'Zlib error';

            case 14:
                return 'Malloc failure';

            case 15:
                return 'Entry has been changed';

            case 16:
                return 'Compression method not supported';

            case 17:
                return 'Premature EOF';

            case 18:
                return 'Invalid argument';

            case 19:
                return 'Not a zip archive';

            case 20:
                return 'Internal error';

            case 21:
                return 'Zip archive inconsistent';

            case 22:
                return 'Can\'t remove file';

            case 23:
                return 'Entry has been deleted';

            default:
                return 'An unknown error has occurred(' . intval($code) . ')';
        }
    }
}


class Chunk
{
    private $_lastChunk = null;
    private $_chunkFolder = '';
    private $_lockFile = '';
    private $_isLast = false;

    public function save($info) {
        if (!$this->validate($info)) {
            return false;
        }

        $this->_lastChunk = $info;
        $this->_chunkFolder = getcwd() . DS . 'upload' . DS . $info['id'];
        $this->_lockFile = $this->_chunkFolder . DS . 'lock';

        if (!is_dir($this->_chunkFolder)) {
            @mkdir($this->_chunkFolder, 0777, true);
        }

        if (!checkDirPermissions($this->_chunkFolder)) {
            throw new PermissionException('Incorrect permissions for ' . $this->_chunkFolder);
        } else {
            $f = fopen($this->_lockFile, 'c');

            if (flock($f, LOCK_EX)) {
                $chunks = array_diff(scandir($this->_chunkFolder), array('.', '..', 'lock'));

                if ((int)$this->_lastChunk['total'] === count($chunks) + 1) {
                    $this->_isLast = true;
                }

                if (!empty($this->_lastChunk['blob'])) {
                    if (empty($_FILES['content']['tmp_name'])) {
                        return false;
                    }

                    move_uploaded_file(
                        $_FILES['content']['tmp_name'],
                        $this->_chunkFolder . DS . (int) $info['current']
                    );
                } else {
                    file_put_contents($this->_chunkFolder . DS . (int) $info['current'], $info['content']);
                }

                flock($f, LOCK_UN);

                return true;
            } else {
                throw new PermissionException('Couldn\'t lock the file');
            }
        }
    }

    public function last() {
        return $this->_isLast;
    }

    public function complete() {
        $content = '';
        for ($i = 1, $count = (int) $this->_lastChunk['total']; $i <= $count; $i++) {
            if (!file_exists($this->_chunkFolder . DS . $i))
                throw new Exception('Missing chunk #' . $i . ' : ' . implode(' / ', scandir($this->_chunkFolder)));

            $data = file_get_contents($this->_chunkFolder . DS . $i);

            if (!empty($this->_lastChunk['encode']) || !empty($this->_lastChunk['zip'])) {
                $data = base64_decode($data);
            }

            $content .= $data;
        }
        rm_r($this->_chunkFolder, [], true);

        if (!empty($this->_lastChunk['zip'])) {
            $result = ZipHelper::unzipString($content, 'data');
        } else if (!empty($this->_lastChunk['encode'])) {
            $result = [
                'status' => 'done',
                'data' => rawurldecode($content)
            ];
        } else {
            $result = [
                'status' => 'done',
                'data' => $content
            ];
        }

        return $result;
    }

    private function validate($info) {
        return !empty($info['id'])      &&
        isset($info['total'])       && (int) $info['total']   >= 1 &&
        isset($info['current'])     && (int) $info['current'] >= 1 &&
        (!empty($_FILES['content']) || !empty($info['content']));
    }

    public static function getInfo() {
        return [
            'id'      => getParam('id', ''),
            'content' => getParam('content', ''),
            'current' => getParam('current', ''),
            'total'   => getParam('total', ''),
            'encode'  => getParam('encode', false),
            'blob'    => getParam('blob', false),
            'zip'     => getParam('zip', false)
        ];
    }

    public static function process($success) {
        $info = self::getInfo();


        $chunk = new self();

        if (!$chunk->save($info)) {
            header($_SERVER['SERVER_PROTOCOL'] . ' 400 Bad Request', true, 400);
            throw new Exception('Chunk save retry');
        }

        if ($chunk->last()) {
            $result = $chunk->complete();
            if ($result['status'] === 'done') {
                $response = $success(json_decode($result['data'], true));
            } else {
                $result['result'] = 'error';
                $response = $result;
            }
        } else {
            $response = ['result' => 'processed'];
        }

        return $response;
    }

}

class PermissionException extends Exception {

}

//